/* 

	parse_config.c
	(C) Nicholas J Humfrey <njh@aelius.com> 2006.
	
	Copyright notice:
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "udpshout.h"




static void process_statement_server( char* name, char* value, int line_num )
{

	if (strcmp( "host", name ) == 0) {
		strncpy( udpshout_server.host, value, STR_BUF_SIZE);
	} else if (strcmp( "port", name ) == 0) {
		udpshout_server.port = atoi( value );
	} else if (strcmp( "user", name ) == 0) { 
		strncpy( udpshout_server.user, value, STR_BUF_SIZE);
	} else if (strcmp( "password", name ) == 0) { 
		strncpy( udpshout_server.password, value, STR_BUF_SIZE);
	} else if (strcmp( "protocol", name ) == 0) {
		if (strcmp( "http", value )==0 || strcmp( "icecast2", value )==0) {
			udpshout_server.protocol = SHOUT_PROTOCOL_HTTP;
		} else if (strcmp( "xaudiocast", value )==0 || strcmp( "icecast1", value )==0) {
			udpshout_server.protocol = SHOUT_PROTOCOL_XAUDIOCAST;
		} else if (strcmp( "icq", value )==0 || strcmp( "shoutcast", value )==0) {
			udpshout_server.protocol = SHOUT_PROTOCOL_ICY;
		} else {
			fprintf(stderr, "Error parsing configuation line %d: invalid protocol.\n", line_num);
			exit(-1);
		}
	} else {
		fprintf(stderr, "Error parsing configuation line %d: invalid statement in section 'server'.\n", line_num);
		exit(-1);
	}
	
}



static void process_statement_channel( char* name, char* value, int line_num )
{
	udpshout_channel_t *chan =  channels[ channel_count-1 ];
	
	if (strcmp( "name", name ) == 0) {
		strncpy( chan->name, value, STR_BUF_SIZE);
		
	} else if (strcmp( "mount", name ) == 0) { 
		strncpy( chan->mount, value, STR_BUF_SIZE);
		
	} else if (strcmp( "pid", name ) == 0) { 
	
		// Check PID is valid
		chan->pid = atoi( value );
		if (chan->pid == 0) {
			fprintf(stderr,"Error parsing configuation line %d: invalid PID\n", line_num);
			exit(-1);
		}
		
		// Add channel to the channel map
		if( channel_map[ chan->pid ] ) {
			fprintf(stderr,"Error parsing configuation line %d: duplicate PID\n", line_num);
			exit(-1);
		} else {
			channel_map[ chan->pid ] = chan;
		}
		
	} else if (strcmp( "genre", name ) == 0) { 
		strncpy( chan->genre, value, STR_BUF_SIZE);
	
	} else if (strcmp( "public", name ) == 0) { 
		chan->is_public = atoi( value );
	
	} else if (strcmp( "description", name ) == 0) { 
		strncpy( chan->description, value, STR_BUF_SIZE);

	} else if (strcmp( "url", name ) == 0) { 
		strncpy( chan->url, value, STR_BUF_SIZE);

	} else if (strcmp( "udp_ip", name ) == 0) { 
		strncpy( chan->udp_ip, value, STR_BUF_SIZE);
		
	} else if (strcmp( "udp_localaddr", name ) == 0) { 
		chan->udp_localaddr = strdup(value);

	} else if (strcmp( "udp_port", name ) == 0) { 
		chan->udp_port = atoi(value);
	} else {
		fprintf(stderr, "Error parsing configuation line %d: invalid statement in section 'channel'.\n", line_num);
		exit(-1);
	}
	
		
}

int parse_config( char *filepath )
{

	FILE* file = NULL;
	char line[STR_BUF_SIZE];
	char section[STR_BUF_SIZE];
	char* ptr;
	int i, line_num=0;
	
	// Initialize strings
	line[0] = '\0';
	section[0] = '\0';
	
	
	// Open the input file
	file = fopen( filepath, "r" );
	if (file==NULL) {
		perror("Failed to open config file");
		exit(-1);
	}
	
	
	// Parse it line by line
	while( !feof( file ) ) {
		line_num++;
		
		fgets( line, STR_BUF_SIZE, file );
		
		// Ignore lines starting with a #
		if (line[0] == '#') continue;
		
		// Remove newline from end
		if (line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = '\0';

		// Ignore empty lines
		if (strlen(line) == 0) continue;
		
		// Is it the start of a section?
		if (line[0]=='[') {
			ptr = &line[1];
			for(i=0; i<strlen(ptr); i++) {
				if (ptr[i] == ']') ptr[i] = '\0';
			}
			
			if (strcmp( ptr, "server")==0) {
				strcpy( section, ptr );
				
			} else if (strcmp( ptr, "channel")==0) {
				udpshout_channel_t* chan=NULL;
				
				if (channel_count >= MAX_CHANNEL_COUNT) {
					fprintf(stderr, "Error: reached maximum number of channels allowed at line %d\n", line_num);
					return 1;
				}
				
				// Allocate memory for channel structure
				chan = calloc( 1, sizeof(udpshout_channel_t) );
				if (!chan) {
					perror("Failed to allocate memory for new channel\n");
					exit(-1);
				}

				chan->num = channel_count;
				chan->shout = NULL;
				chan->pes_stream_id = 0;
				chan->udpsockfd = -1;
				chan->continuity_count = -1;
				chan->pes_ts = 0;
				
				chan->udp_port = 1234;
				chan->udp_localaddr = NULL;

				
				strcpy( section, ptr );
				channels[ channel_count ] = chan;
				channel_count++;
				
			} else {
				fprintf(stderr, "Error parsing configuation line %d: unknown section '%s'\n", line_num, ptr);
				exit(-1);
			}
			
			
			
		} else {
			char* name = line;
			char* value = NULL;
			
			// Split up the name and value
			for(i=0; i<strlen(name); i++) {
				if (name[i] == ':') {
					name[i] = '\0';
					value = &name[i+1];
					if (value[0] == ' ') { value++; }
					if (strlen(value)==0) { value=NULL; }
				}
			}
			
			// Ignore empty values
			if (value==NULL) continue;
			
			if (strcmp( section, "server")==0) {	
				process_statement_server( name, value, line_num );
			} else if (strcmp( section, "channel")==0) {	
				process_statement_channel( name, value, line_num );
			} else {
				fprintf(stderr, "Error parsing configuation line %d: missing section.\n", line_num);
				exit(-1);
			}

		}
		
		
	}
	

	fclose( file );
	
	return 0;
}

