
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <pthread.h>


#include "udpshout.h"
#include "config.h"



int Interrupted=0;

int channel_count=0;
udpshout_channel_t *channels[MAX_CHANNEL_COUNT];
udpshout_channel_t *channel_map[MAX_PID_COUNT];
udpshout_server_t udpshout_server;


static void signal_handler(int signum)
{
	if (signum != SIGPIPE) {
		Interrupted=signum;
	}
	signal(signum,signal_handler);
}



static shout_t * connect_server_channel( udpshout_channel_t *chan )
{
	shout_t *shout;
	char string[STR_BUF_SIZE];
	int result;

	// Don't connect?
	if (strlen(udpshout_server.host)==0) {
		return NULL;
	} else {
		shout = shout_new();
	}


	// Set server parameters
	shout_set_agent( shout, PACKAGE_STRING );
	shout_set_host( shout, udpshout_server.host );
	shout_set_port( shout, udpshout_server.port );
	shout_set_user( shout, udpshout_server.user );
	shout_set_password( shout, udpshout_server.password );
	shout_set_protocol( shout, udpshout_server.protocol );
	shout_set_format( shout, SHOUT_FORMAT_MP3 );

	shout_set_name( shout, chan->name );
	shout_set_mount( shout, chan->mount );
	shout_set_genre( shout, chan->genre );
	shout_set_description( shout, chan->description );
	shout_set_url( shout, chan->url );


	// Add information about the audio format
	snprintf(string, STR_BUF_SIZE, "%d", chan->mpah.bitrate );
	shout_set_audio_info( shout, SHOUT_AI_BITRATE, string);
	
	snprintf(string, STR_BUF_SIZE, "%d", chan->mpah.samplerate );
	shout_set_audio_info( shout, SHOUT_AI_SAMPLERATE, string);
	
	snprintf(string, STR_BUF_SIZE, "%d", chan->mpah.channels );
	shout_set_audio_info( shout, SHOUT_AI_CHANNELS, string);
	
	
	// Connect!
	fprintf(stderr, "  Connecting to: http://%s:%d%s\n",
		shout_get_host( shout ), shout_get_port( shout ), shout_get_mount( shout ));
		
	result = shout_open( shout );
	if (result != SHOUTERR_SUCCESS) {
		fprintf(stderr,"  Failed to connect to server: %s.\n", shout_get_error(shout));
		exit(-1);
	}
	
	
	return shout;
}

static void parse_args(int argc, char **argv) 
{
	if (argc>1) {
	
 		parse_config( argv[1] );

 	} else {
		fprintf(stderr,"%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
		fprintf(stderr,"Usage: udpshout <configfile>\n\n");
		exit(-1);
	}
}

static void ts_continuity_check( udpshout_channel_t *chan, int ts_cc ) 
{
	if (chan->continuity_count != ts_cc) {
	
		// Only display an error after we gain sync
		if (chan->synced) {
			fprintf(stderr, "Warning: TS continuity error (pid: %d)\n", chan->pid );
			chan->synced=0;
		}
		chan->continuity_count = ts_cc;
	}

	chan->continuity_count++;
	if (chan->continuity_count==16)
		chan->continuity_count=0;
}

static void extract_pes_payload( unsigned char *pes_ptr, size_t pes_len, udpshout_channel_t *chan, int start_of_pes ) 
{
	unsigned char* es_ptr=NULL;
	size_t es_len=0;
	
	
	// Start of a PES header?
	if ( start_of_pes ) {
		// Parse the PES header
		es_ptr = parse_pes( pes_ptr, pes_len, &es_len, chan);
	} else if (chan->pes_stream_id) {
	
		// Don't output any data until we have seen a PES header
		es_ptr = pes_ptr;
		es_len = pes_len;
	
		// Are we are the end of the PES packet?
		if (es_len>chan->pes_remaining) {
			es_len=chan->pes_remaining;
		}
		
	}
	
	// Subtract the amount remaining in current PES packet
	chan->pes_remaining -= es_len;

	// Got some data to write out?
	if (es_ptr) {
	
		// Scan through Elementary Stream (ES) 
		// and try and find MPEG audio stream header
		while (!chan->synced && es_len>=4) {
		
			// Valid header?
			if (mpa_header_parse(es_ptr, &chan->mpah)) {

				// Now we know bitrate etc, set things up
				fprintf(stderr, "Synced to MPEG audio for '%s' (pid: %d, stream: 0x%x payload: %d)\n",  chan->name, chan->pid, chan->pes_stream_id, chan->mpah.framesize );
				mpa_header_print( &chan->mpah );
				chan->synced = 1;
				
				chan->payload_size = chan->mpah.framesize;
				
				// Allocate buffer to store packet in
				chan->buf_size = chan->payload_size + TS_PACKET_SIZE;
				chan->buf = realloc( chan->buf, chan->buf_size + 4 );
				if (chan->buf==NULL) {
					perror("Error: Failed to allocate memory for MPEG Audio buffer");
					exit(-1);
				}

				// (re-)connect to the server
				if (!chan->shout)
					chan->shout = connect_server_channel( chan );

				fprintf(stderr, "\n");
				
			} else {
				// Skip byte
				es_len--;
				es_ptr++;
			}
		}
		
		
		// If stream is synced then put data info buffer
		if (chan->synced) {
		
			// Check that there is space
			if (chan->buf_used + es_len > chan->buf_size) {
				fprintf(stderr, "Error: MPEG Audio buffer overflow\n" );
				exit(-1);
			}
		
			// Copy data into the buffer
			memcpy( chan->buf + chan->buf_used, es_ptr, es_len);
			chan->buf_used += es_len;
		}
	}
	
	
	// Got enough to send packet?
	if (chan->buf_used > chan->payload_size ) {
		int result;
		int frame_size_with_padding = chan->payload_size + ((chan->buf[2] >> 1) & 0x01);
		
		// Ensure a MPEG Audio frame starts here
		if (chan->buf[0] != 0xFF) {
			fprintf(stderr, "Warning: lost MPEG Audio sync for PID %d.\n", chan->pid);
			chan->synced = 0;
			chan->buf_used = 0;
			return;
		}
	
		// Send the data to the shoutcast server
		if ( chan->shout ) {
			result = shout_send_raw( chan->shout, chan->buf, frame_size_with_padding);
			if (result < 0) {
				fprintf(stderr, "Error: failed to send data to server for PID %d.\n", chan->pid);
				fprintf(stderr, "  libshout: %s.\n", shout_get_error(chan->shout));
				exit(-1);
			}
		}
		
		// Move any remaining memory to the start of the buffer
		chan->buf_used -= frame_size_with_padding;
		memmove( chan->buf, chan->buf + frame_size_with_padding, chan->buf_used );
		
	}
}



static void *process_ts_packets(void *vptr_args)
{
	udpshout_channel_t *chan = vptr_args;
//	unsigned char buf[TS_PACKET_SIZE];
	unsigned char buf[1500];
	unsigned char *buf_ptr = buf;
	unsigned char* pes_ptr=NULL;
	unsigned int pid=0;
	size_t pes_len;
	int bytes_read = 0;
	
	
	if (!input_init_udp(chan))
		pthread_exit(NULL);
	

	while ( !Interrupted ) {
		/* do we have anough buffer to parse one TS packet */
		if (buf_ptr + TS_PACKET_SIZE < buf+bytes_read) {
			buf_ptr += TS_PACKET_SIZE;
		} else {
			chan->last_update=time(NULL);
			bytes_read = read(chan->udpsockfd,buf,1500);
			if (bytes_read==0) continue;
			if (bytes_read % TS_PACKET_SIZE != 0) {
				fprintf(stderr, "Thread [%d] Error: Unknown TS packets count in udp datagram "
								"(%d bytes) may be %d packets\n", chan->num,
								bytes_read,  (bytes_read/TS_PACKET_SIZE));
				break;
			}
			buf_ptr = buf;
		}
		
		// Check the sync-byte
		if (TS_PACKET_SYNC_BYTE(buf_ptr) != 0x47) {
			fprintf(stderr,"Thread [%d] Error: Lost syncronisation - aborting\n", chan->num);
			break;
		}
		
		// Transport error?
		if ( TS_PACKET_TRANS_ERROR(buf_ptr) ) {
			fprintf(stderr, "Thread [%d] Warning: Transport error in PID %d.\n", chan->num, pid);
			if (channel_map[ pid ]) {
				channel_map[ pid ]->synced = 0;
				channel_map[ pid ]->buf_used = 0;
			}
			continue;
		}			

		// Get the PID of this TS packet
		pid = TS_PACKET_PID(buf_ptr);

		// Check we know about the payload
		if (channel_map[ pid ]) {

		    // Scrambled?
		    if ( TS_PACKET_SCRAMBLING(buf_ptr) ) {
			fprintf(stderr, "Thread [%d] Warning: PID %d is scrambled.\n", chan->num, pid);
			continue;
		    }	

		    // Location of and size of PES payload
		    pes_ptr = &buf_ptr[4];
		    pes_len = TS_PACKET_SIZE - 4;

		    // Check for adaptation field?
		    if (TS_PACKET_ADAPTATION(buf_ptr)==0x1) {
			// Payload only, no adaptation field
		    } else if (TS_PACKET_ADAPTATION(buf_ptr)==0x2) {
			// Adaptation field only, no payload
			continue;
		    } else if (TS_PACKET_ADAPTATION(buf_ptr)==0x3) {
			// Adaptation field AND payload
			pes_ptr += (TS_PACKET_ADAPT_LEN(buf_ptr) + 1);
			pes_len -= (TS_PACKET_ADAPT_LEN(buf_ptr) + 1);
		    }
		    
		    // Continuity check
		    ts_continuity_check( channel_map[ pid ], TS_PACKET_CONT_COUNT(buf_ptr) );

		    // Extract PES payload and put it in FIFO
		    extract_pes_payload( pes_ptr, pes_len, channel_map[ pid ], TS_PACKET_PAYLOAD_START(buf_ptr) );	
		} else {
			//fprintf(stderr, "Thread [%d] Warning: don't know anything about PID %d.\n", chan->num, pid);
		}
	}
	pthread_exit(NULL);
}



int main(int argc, char **argv)
{
	int i;
	pthread_t threads[MAX_CHANNEL_COUNT];
	
	for (i=0;i<MAX_PID_COUNT;i++) channel_map[i]=NULL;
	for (i=0;i<MAX_CHANNEL_COUNT;i++) channels[i]=NULL;
	memset( &udpshout_server, 0, sizeof(udpshout_server_t) );
	
	// Default server settings
	udpshout_server.port = SERVER_PORT_DEFAULT;
	strcpy(udpshout_server.user, SERVER_USER_DEFAULT);
	strcpy(udpshout_server.password, SERVER_PASSWORD_DEFAULT);
	udpshout_server.protocol = SERVER_PROTOCOL_DEFAULT;
	
	// Initialise libshout
	shout_init();

	// Parse command line arguments
	parse_args( argc, argv );
	
	// Setup signal handlers
	if (signal(SIGHUP, signal_handler) == SIG_IGN) signal(SIGHUP, SIG_IGN);
	if (signal(SIGINT, signal_handler) == SIG_IGN) signal(SIGINT, SIG_IGN);
	if (signal(SIGTERM, signal_handler) == SIG_IGN) signal(SIGTERM, SIG_IGN);


	// Read and process TS packets from DVR device
	for (i=0;i<channel_count;i++) {
		if (pthread_create(&threads[i], NULL, process_ts_packets, (void *) channels[i]) != 0){
			fprintf(stderr,"init thread %i: ",i);
			perror("Failed to create thread");
			return -1;
		}
		fprintf(stderr,"thread [%d] - created\n",i);
	}
	
	
	// watchdog
	while(!Interrupted) {
		time_t now_time = time(NULL);
		for (i=0;i<channel_count;i++)
			if (now_time > channels[i]->last_update+15) {
				fprintf(stderr, "Cant find stream on UDP: %s:%d\n",
									channels[i]->udp_ip, channels[i]->udp_port);
			/* do something here to repair stream
			 * maybe sending detach/join again can help */
			}
		sleep(10);
	}
	

	if (Interrupted) {
		fprintf(stderr,"Caught signal %d - closing cleanly.\n",Interrupted);
	}
	
	
	//cleaning (close fd's)
	
	// Clean up
	for (i=0;i<channel_count;i++) {
		if (channels[i]->udpsockfd != -1) socket_close(channels[i]->udpsockfd);
		if (channels[i]->buf) free( channels[i]->buf );
		if (channels[i]->shout) {
			shout_close( channels[i]->shout );
			shout_free( channels[i]->shout );
		}
		free( channels[i] );
	}
	// Shutdown libshout
	shout_shutdown();	

	return(0);
}

